"""OpenAPI core protocols module"""
import sys
from typing import Any
from typing import Mapping
from typing import Optional

if sys.version_info >= (3, 8):
    from typing import Protocol
    from typing import runtime_checkable
else:
    from typing_extensions import Protocol
    from typing_extensions import runtime_checkable

from openapi_core.datatypes import RequestParameters


@runtime_checkable
class BaseRequest(Protocol):
    parameters: RequestParameters

    @property
    def method(self) -> str:
        ...

    @property
    def body(self) -> Optional[str]:
        ...

    @property
    def mimetype(self) -> str:
        ...


@runtime_checkable
class Request(BaseRequest, Protocol):
    """Request attributes protocol.

    Attributes:
        host_url
            Url with scheme and host
            For example:
            https://localhost:8000
        path
            Request path
        full_url_pattern
            The matched url with scheme, host and path pattern.
            For example:
            https://localhost:8000/api/v1/pets
            https://localhost:8000/api/v1/pets/{pet_id}
        method
            The request method, as lowercase string.
        parameters
            A RequestParameters object. Needs to supports path attribute setter
            to write resolved path parameters.
        body
            The request body, as string.
        mimetype
            Like content type, but without parameters (eg, without charset,
            type etc.) and always lowercase.
            For example if the content type is "text/HTML; charset=utf-8"
            the mimetype would be "text/html".
    """

    @property
    def host_url(self) -> str:
        ...

    @property
    def path(self) -> str:
        ...


@runtime_checkable
class WebhookRequest(BaseRequest, Protocol):
    """Webhook request attributes protocol.

    Attributes:
        name
            Webhook name
        method
            The request method, as lowercase string.
        parameters
            A RequestParameters object. Needs to supports path attribute setter
            to write resolved path parameters.
        body
            The request body, as string.
        mimetype
            Like content type, but without parameters (eg, without charset,
            type etc.) and always lowercase.
            For example if the content type is "text/HTML; charset=utf-8"
            the mimetype would be "text/html".
    """

    @property
    def name(self) -> str:
        ...


@runtime_checkable
class SupportsPathPattern(Protocol):
    """Supports path_pattern attribute protocol.

    You also need to provide path variables in RequestParameters.

    Attributes:
        path_pattern
            The matched path pattern.
            For example:
            /api/v1/pets/{pet_id}
    """

    @property
    def path_pattern(self) -> str:
        ...


@runtime_checkable
class Response(Protocol):
    """Response protocol.

    Attributes:
        data
            The response body, as string.
        status_code
            The status code as integer.
        headers
            Response headers as Headers.
        mimetype
            Lowercase content type without charset.
    """

    @property
    def data(self) -> str:
        ...

    @property
    def status_code(self) -> int:
        ...

    @property
    def mimetype(self) -> str:
        ...

    @property
    def headers(self) -> Mapping[str, Any]:
        ...



from openapi_core.schema.servers import get_server_url
from openapi_core.spec import Spec


def get_spec_url(spec: Spec, index: int = 0) -> str:
    servers = spec / "servers"
    return get_server_url(servers / 0)



"""OpenAPI core contrib falcon responses module"""
from falcon.response import Response
from werkzeug.datastructures import Headers


class FalconOpenAPIResponse:
    def __init__(self, response: Response):
        if not isinstance(response, Response):
            raise TypeError(f"'response' argument is not type of {Response}")
        self.response = response

    @property
    def data(self) -> str:
        if self.response.text is None:
            return ""
        assert isinstance(self.response.text, str)
        return self.response.text

    @property
    def status_code(self) -> int:
        return int(self.response.status[:3])

    @property
    def mimetype(self) -> str:
        mimetype = ""
        if self.response.content_type:
            mimetype = self.response.content_type.partition(";")[0]
        else:
            mimetype = self.response.options.default_media_type
        return mimetype

    @property
    def headers(self) -> Headers:
        return Headers(self.response.headers)



from typing import Any
from typing import Mapping
from typing import Optional
from typing import Tuple

from openapi_core.casting.schemas import schema_casters_factory
from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.deserializing.media_types import (
    media_type_deserializers_factory,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.parameters import (
    parameter_deserializers_factory,
)
from openapi_core.deserializing.parameters.factories import (
    ParameterDeserializersFactory,
)
from openapi_core.spec import Spec
from openapi_core.unmarshalling.schemas.datatypes import (
    FormatUnmarshallersDict,
)
from openapi_core.unmarshalling.schemas.factories import (
    SchemaUnmarshallersFactory,
)
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory
from openapi_core.validation.validators import BaseValidator


class BaseUnmarshaller(BaseValidator):
    schema_unmarshallers_factory: SchemaUnmarshallersFactory = NotImplemented

    def __init__(
        self,
        spec: Spec,
        base_url: Optional[str] = None,
        schema_casters_factory: SchemaCastersFactory = schema_casters_factory,
        parameter_deserializers_factory: ParameterDeserializersFactory = parameter_deserializers_factory,
        media_type_deserializers_factory: MediaTypeDeserializersFactory = media_type_deserializers_factory,
        schema_validators_factory: Optional[SchemaValidatorsFactory] = None,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
        schema_unmarshallers_factory: Optional[
            SchemaUnmarshallersFactory
        ] = None,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
    ):
        if schema_validators_factory is None and schema_unmarshallers_factory:
            schema_validators_factory = (
                schema_unmarshallers_factory.schema_validators_factory
            )
        super().__init__(
            spec,
            base_url=base_url,
            schema_casters_factory=schema_casters_factory,
            parameter_deserializers_factory=parameter_deserializers_factory,
            media_type_deserializers_factory=media_type_deserializers_factory,
            schema_validators_factory=schema_validators_factory,
            format_validators=format_validators,
            extra_format_validators=extra_format_validators,
            extra_media_type_deserializers=extra_media_type_deserializers,
        )
        self.schema_unmarshallers_factory = (
            schema_unmarshallers_factory or self.schema_unmarshallers_factory
        )
        if self.schema_unmarshallers_factory is NotImplemented:
            raise NotImplementedError(
                "schema_unmarshallers_factory is not assigned"
            )
        self.format_unmarshallers = format_unmarshallers
        self.extra_format_unmarshallers = extra_format_unmarshallers

    def _unmarshal_schema(self, schema: Spec, value: Any) -> Any:
        unmarshaller = self.schema_unmarshallers_factory.create(
            schema,
            format_validators=self.format_validators,
            extra_format_validators=self.extra_format_validators,
            format_unmarshallers=self.format_unmarshallers,
            extra_format_unmarshallers=self.extra_format_unmarshallers,
        )
        return unmarshaller.unmarshal(value)

    def _get_param_or_header_value(
        self,
        param_or_header: Spec,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Any:
        casted, schema = self._get_param_or_header_value_and_schema(
            param_or_header, location, name
        )
        if schema is None:
            return casted
        return self._unmarshal_schema(schema, casted)

    def _get_content_value(
        self, raw: Any, mimetype: str, content: Spec
    ) -> Any:
        casted, schema = self._get_content_value_and_schema(
            raw, mimetype, content
        )
        if schema is None:
            return casted
        return self._unmarshal_schema(schema, casted)



import sys

if sys.version_info >= (3, 8):
    from typing import Protocol
    from typing import runtime_checkable
else:
    from typing_extensions import Protocol
    from typing_extensions import runtime_checkable

from requests.cookies import RequestsCookieJar


@runtime_checkable
class SupportsCookieJar(Protocol):
    _cookies: RequestsCookieJar



"""OpenAPI core contrib django responses module"""
from django.http.response import HttpResponse
from werkzeug.datastructures import Headers


class DjangoOpenAPIResponse:
    def __init__(self, response: HttpResponse):
        if not isinstance(response, HttpResponse):
            raise TypeError(
                f"'response' argument is not type of {HttpResponse}"
            )
        self.response = response

    @property
    def data(self) -> str:
        assert isinstance(self.response.content, bytes)
        return self.response.content.decode("utf-8")

    @property
    def status_code(self) -> int:
        assert isinstance(self.response.status_code, int)
        return self.response.status_code

    @property
    def headers(self) -> Headers:
        return Headers(self.response.headers.items())

    @property
    def mimetype(self) -> str:
        content_type = self.response.get("Content-Type", "")
        assert isinstance(content_type, str)
        return content_type



from json import loads

from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.media_types.util import data_form_loads
from openapi_core.deserializing.media_types.util import urlencoded_form_loads

__all__ = ["media_type_deserializers_factory"]

media_type_deserializers: MediaTypeDeserializersDict = {
    "application/json": loads,
    "application/x-www-form-urlencoded": urlencoded_form_loads,
    "multipart/form-data": data_form_loads,
}

media_type_deserializers_factory = MediaTypeDeserializersFactory(
    media_type_deserializers=media_type_deserializers,
)



"""OpenAPI core contrib starlette requests module"""
from typing import Optional

from asgiref.sync import AsyncToSync
from starlette.requests import Request

from openapi_core.datatypes import RequestParameters


class StarletteOpenAPIRequest:
    def __init__(self, request: Request):
        if not isinstance(request, Request):
            raise TypeError(f"'request' argument is not type of {Request}")
        self.request = request

        self.parameters = RequestParameters(
            query=self.request.query_params,
            header=self.request.headers,
            cookie=self.request.cookies,
        )

        self._get_body = AsyncToSync(self.request.body, force_new_loop=True)  # type: ignore

    @property
    def host_url(self) -> str:
        return self.request.base_url._url

    @property
    def path(self) -> str:
        return self.request.url.path

    @property
    def method(self) -> str:
        return self.request.method.lower()

    @property
    def body(self) -> Optional[str]:
        body = self._get_body()
        if body is None:
            return None
        if isinstance(body, bytes):
            return body.decode("utf-8")
        assert isinstance(body, str)
        return body

    @property
    def mimetype(self) -> str:
        content_type = self.request.headers["Content-Type"]
        if content_type:
            return content_type.partition(";")[0]

        return ""



"""OpenAPI core contrib starlette responses module"""
from starlette.datastructures import Headers
from starlette.responses import Response


class StarletteOpenAPIResponse:
    def __init__(self, response: Response):
        if not isinstance(response, Response):
            raise TypeError(f"'response' argument is not type of {Response}")
        self.response = response

    @property
    def data(self) -> str:
        if isinstance(self.response.body, bytes):
            return self.response.body.decode("utf-8")
        assert isinstance(self.response.body, str)
        return self.response.body

    @property
    def status_code(self) -> int:
        return self.response.status_code

    @property
    def mimetype(self) -> str:
        return self.response.media_type or ""

    @property
    def headers(self) -> Headers:
        return self.response.headers



import re
from typing import Any
from typing import Dict
from typing import Mapping
from typing import Optional

from openapi_core.schema.protocols import SuportsGetAll
from openapi_core.schema.protocols import SuportsGetList
from openapi_core.spec import Spec


def get_aslist(param_or_header: Spec) -> bool:
    """Checks if parameter/header is described as list for simpler scenarios"""
    # if schema is not defined it's a complex scenario
    if "schema" not in param_or_header:
        return False

    schema = param_or_header / "schema"
    schema_type = schema.getkey("type", "any")
    # TODO: resolve for 'any' schema type
    return schema_type in ["array", "object"]


def get_style(param_or_header: Spec) -> str:
    """Checks parameter/header style for simpler scenarios"""
    if "style" in param_or_header:
        assert isinstance(param_or_header["style"], str)
        return param_or_header["style"]

    # if "in" not defined then it's a Header
    location = param_or_header.getkey("in", "header")

    # determine default
    return "simple" if location in ["path", "header"] else "form"


def get_explode(param_or_header: Spec) -> bool:
    """Checks parameter/header explode for simpler scenarios"""
    if "explode" in param_or_header:
        assert isinstance(param_or_header["explode"], bool)
        return param_or_header["explode"]

    # determine default
    style = get_style(param_or_header)
    return style == "form"


def get_value(
    param_or_header: Spec,
    location: Mapping[str, Any],
    name: Optional[str] = None,
) -> Any:
    """Returns parameter/header value from specific location"""
    name = name or param_or_header["name"]
    style = get_style(param_or_header)

    if name not in location:
        # Only check if the name is not in the location if the style of
        # the param is deepObject,this is because deepObjects will never be found
        # as their key also includes the properties of the object already.
        if style != "deepObject":
            raise KeyError
        keys_str = " ".join(location.keys())
        if not re.search(rf"{name}\[\w+\]", keys_str):
            raise KeyError

    aslist = get_aslist(param_or_header)
    explode = get_explode(param_or_header)
    if aslist and explode:
        if style == "deepObject":
            return get_deep_object_value(location, name)
        if isinstance(location, SuportsGetAll):
            return location.getall(name)
        if isinstance(location, SuportsGetList):
            return location.getlist(name)

    return location[name]


def get_deep_object_value(
    location: Mapping[str, Any],
    name: Optional[str] = None,
) -> Dict[str, Any]:
    values = {}
    for key, value in location.items():
        # Split the key from the brackets.
        key_split = re.split(pattern=r"\[|\]", string=key)
        if key_split[0] == name:
            values[key_split[1]] = value
    return values



from dataclasses import dataclass

from openapi_core.exceptions import OpenAPIError


class UnmarshalError(OpenAPIError):
    """Schema unmarshal operation error"""


class UnmarshallerError(UnmarshalError):
    """Unmarshaller error"""


@dataclass
class FormatterNotFoundError(UnmarshallerError):
    """Formatter not found to unmarshal"""

    type_format: str

    def __str__(self) -> str:
        return f"Formatter not found for {self.type_format} format"


class FormatUnmarshalError(UnmarshallerError):
    """Unable to unmarshal value for format"""

    value: str
    type: str
    original_exception: Exception

    def __str__(self) -> str:
        return (
            "Unable to unmarshal value {value} for format {type}: {exception}"
        ).format(
            value=self.value,
            type=self.type,
            exception=self.original_exception,
        )



"""OpenAPI core unmarshalling request module"""
from openapi_core.unmarshalling.request.proxies import (
    DetectSpecRequestValidatorProxy,
)
from openapi_core.unmarshalling.request.proxies import (
    SpecRequestValidatorProxy,
)
from openapi_core.unmarshalling.request.unmarshallers import (
    APICallRequestUnmarshaller,
)
from openapi_core.unmarshalling.request.unmarshallers import RequestValidator
from openapi_core.unmarshalling.request.unmarshallers import (
    V30RequestUnmarshaller,
)
from openapi_core.unmarshalling.request.unmarshallers import (
    V31RequestUnmarshaller,
)
from openapi_core.unmarshalling.request.unmarshallers import (
    V31WebhookRequestUnmarshaller,
)
from openapi_core.unmarshalling.schemas import (
    oas30_write_schema_unmarshallers_factory,
)
from openapi_core.unmarshalling.schemas import (
    oas31_schema_unmarshallers_factory,
)

__all__ = [
    "V30RequestUnmarshaller",
    "V31RequestUnmarshaller",
    "V31WebhookRequestUnmarshaller",
    "RequestValidator",
    "openapi_v30_request_validator",
    "openapi_v31_request_validator",
    "openapi_v3_request_validator",
    "openapi_request_validator",
]

# alias to the latest v3 version
V3RequestUnmarshaller = V31RequestUnmarshaller
V3WebhookRequestUnmarshaller = V31WebhookRequestUnmarshaller

# spec validators
openapi_v30_request_validator = SpecRequestValidatorProxy(
    APICallRequestUnmarshaller,
    schema_unmarshallers_factory=oas30_write_schema_unmarshallers_factory,
    deprecated="openapi_v30_request_validator",
    use="V30RequestValidator",
)
openapi_v31_request_validator = SpecRequestValidatorProxy(
    APICallRequestUnmarshaller,
    schema_unmarshallers_factory=oas31_schema_unmarshallers_factory,
    deprecated="openapi_v31_request_validator",
    use="V31RequestValidator",
)

# spec validators alias to the latest v3 version
openapi_v3_request_validator = openapi_v31_request_validator

# detect version spec
openapi_request_validator = DetectSpecRequestValidatorProxy(
    {
        ("openapi", "3.0"): openapi_v30_request_validator,
        ("openapi", "3.1"): openapi_v31_request_validator,
    },
)



from typing import Dict

from openapi_core.casting.schemas.casters import ArrayCaster
from openapi_core.casting.schemas.casters import BaseSchemaCaster
from openapi_core.casting.schemas.casters import CallableSchemaCaster
from openapi_core.casting.schemas.casters import DummyCaster
from openapi_core.casting.schemas.datatypes import CasterCallable
from openapi_core.spec import Spec
from openapi_core.util import forcebool


class SchemaCastersFactory:
    DUMMY_CASTERS = [
        "string",
        "object",
        "any",
    ]
    PRIMITIVE_CASTERS: Dict[str, CasterCallable] = {
        "integer": int,
        "number": float,
        "boolean": forcebool,
    }
    COMPLEX_CASTERS = {
        "array": ArrayCaster,
    }

    def create(self, schema: Spec) -> BaseSchemaCaster:
        schema_type = schema.getkey("type", "any")

        if schema_type in self.DUMMY_CASTERS:
            return DummyCaster(schema)

        if schema_type in self.PRIMITIVE_CASTERS:
            caster_callable = self.PRIMITIVE_CASTERS[schema_type]
            return CallableSchemaCaster(schema, caster_callable)

        return ArrayCaster(schema, self)



"""OpenAPI core contrib flask views module"""
from typing import Any

from flask.views import MethodView

from openapi_core.contrib.flask.decorators import FlaskOpenAPIViewDecorator
from openapi_core.contrib.flask.handlers import FlaskOpenAPIErrorsHandler
from openapi_core.spec import Spec


class FlaskOpenAPIView(MethodView):
    """Brings OpenAPI specification validation and unmarshalling for views."""

    openapi_errors_handler = FlaskOpenAPIErrorsHandler

    def __init__(self, spec: Spec):
        super().__init__()
        self.spec = spec

    def dispatch_request(self, *args: Any, **kwargs: Any) -> Any:
        decorator = FlaskOpenAPIViewDecorator(
            self.spec,
            openapi_errors_handler=self.openapi_errors_handler,
        )
        return decorator(super().dispatch_request)(*args, **kwargs)



import logging
import warnings
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Type
from typing import Union

from openapi_core.extensions.models.factories import ModelPathFactory
from openapi_core.schema.schemas import get_properties
from openapi_core.spec import Spec
from openapi_core.unmarshalling.schemas.datatypes import FormatUnmarshaller
from openapi_core.unmarshalling.schemas.datatypes import (
    FormatUnmarshallersDict,
)
from openapi_core.unmarshalling.schemas.exceptions import FormatUnmarshalError
from openapi_core.unmarshalling.schemas.exceptions import UnmarshallerError
from openapi_core.validation.schemas.datatypes import CustomFormattersDict
from openapi_core.validation.schemas.validators import SchemaValidator

log = logging.getLogger(__name__)


class PrimitiveUnmarshaller:
    def __init__(
        self,
        schema: Spec,
        schema_validator: SchemaValidator,
        schema_unmarshaller: "SchemaUnmarshaller",
    ) -> None:
        self.schema = schema
        self.schema_validator = schema_validator
        self.schema_unmarshaller = schema_unmarshaller

    def __call__(self, value: Any) -> Any:
        return value


class ArrayUnmarshaller(PrimitiveUnmarshaller):
    def __call__(self, value: Any) -> Optional[List[Any]]:
        return list(map(self.items_unmarshaller.unmarshal, value))

    @property
    def items_unmarshaller(self) -> "SchemaUnmarshaller":
        # sometimes we don't have any schema i.e. free-form objects
        items_schema = self.schema.get(
            "items", Spec.from_dict({}, validator=None)
        )
        return self.schema_unmarshaller.evolve(items_schema)


class ObjectUnmarshaller(PrimitiveUnmarshaller):
    def __call__(self, value: Any) -> Any:
        properties = self._unmarshal_properties(value)

        fields: Iterable[str] = properties and properties.keys() or []
        object_class = self.object_class_factory.create(self.schema, fields)

        return object_class(**properties)

    @property
    def object_class_factory(self) -> ModelPathFactory:
        return ModelPathFactory()

    def evolve(self, schema: Spec) -> "ObjectUnmarshaller":
        cls = self.__class__

        return cls(
            schema,
            self.schema_validator.evolve(schema),
            self.schema_unmarshaller,
        )

    def _unmarshal_properties(
        self, value: Any, schema_only: bool = False
    ) -> Any:
        properties = {}

        one_of_schema = self.schema_validator.get_one_of_schema(value)
        if one_of_schema is not None:
            one_of_properties = self.evolve(
                one_of_schema
            )._unmarshal_properties(value, schema_only=True)
            properties.update(one_of_properties)

        any_of_schemas = self.schema_validator.iter_any_of_schemas(value)
        for any_of_schema in any_of_schemas:
            any_of_properties = self.evolve(
                any_of_schema
            )._unmarshal_properties(value, schema_only=True)
            properties.update(any_of_properties)

        all_of_schemas = self.schema_validator.iter_all_of_schemas(value)
        for all_of_schema in all_of_schemas:
            all_of_properties = self.evolve(
                all_of_schema
            )._unmarshal_properties(value, schema_only=True)
            properties.update(all_of_properties)

        for prop_name, prop_schema in get_properties(self.schema).items():
            try:
                prop_value = value[prop_name]
            except KeyError:
                if "default" not in prop_schema:
                    continue
                prop_value = prop_schema["default"]

            properties[prop_name] = self.schema_unmarshaller.evolve(
                prop_schema
            ).unmarshal(prop_value)

        if schema_only:
            return properties

        additional_properties = self.schema.getkey(
            "additionalProperties", True
        )
        if additional_properties is not False:
            # free-form object
            if additional_properties is True:
                additional_prop_schema = Spec.from_dict(
                    {"nullable": True}, validator=None
                )
            # defined schema
            else:
                additional_prop_schema = self.schema / "additionalProperties"
            additional_prop_unmarshaler = self.schema_unmarshaller.evolve(
                additional_prop_schema
            )
            for prop_name, prop_value in value.items():
                if prop_name in properties:
                    continue
                properties[prop_name] = additional_prop_unmarshaler.unmarshal(
                    prop_value
                )

        return properties


class MultiTypeUnmarshaller(PrimitiveUnmarshaller):
    def __call__(self, value: Any) -> Any:
        unmarshaller = self._get_best_unmarshaller(value)
        return unmarshaller(value)

    @property
    def type(self) -> List[str]:
        types = self.schema.getkey("type", ["any"])
        assert isinstance(types, list)
        return types

    def _get_best_unmarshaller(self, value: Any) -> "PrimitiveUnmarshaller":
        for schema_type in self.type:
            result = self.schema_validator.type_validator(
                value, type_override=schema_type
            )
            if not result:
                continue
            result = self.schema_validator.format_validator(value)
            if not result:
                continue
            return self.schema_unmarshaller.get_type_unmarshaller(schema_type)

        raise UnmarshallerError("Unmarshaller not found for type(s)")


class AnyUnmarshaller(MultiTypeUnmarshaller):
    SCHEMA_TYPES_ORDER = [
        "object",
        "array",
        "boolean",
        "integer",
        "number",
        "string",
    ]

    @property
    def type(self) -> List[str]:
        return self.SCHEMA_TYPES_ORDER


class TypesUnmarshaller:
    unmarshallers: Mapping[str, Type[PrimitiveUnmarshaller]] = {}
    multi: Optional[Type[PrimitiveUnmarshaller]] = None

    def __init__(
        self,
        unmarshallers: Mapping[str, Type[PrimitiveUnmarshaller]],
        default: Type[PrimitiveUnmarshaller],
        multi: Optional[Type[PrimitiveUnmarshaller]] = None,
    ):
        self.unmarshallers = unmarshallers
        self.default = default
        self.multi = multi

    def get_unmarshaller(
        self,
        schema_type: Optional[Union[Iterable[str], str]],
    ) -> Type["PrimitiveUnmarshaller"]:
        if schema_type is None:
            return self.default
        if isinstance(schema_type, Iterable) and not isinstance(
            schema_type, str
        ):
            if self.multi is None:
                raise TypeError("Unmarshaller does not accept multiple types")
            return self.multi

        return self.unmarshallers[schema_type]


class FormatsUnmarshaller:
    def __init__(
        self,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        custom_formatters: Optional[CustomFormattersDict] = None,
    ):
        if format_unmarshallers is None:
            format_unmarshallers = {}
        self.format_unmarshallers = format_unmarshallers
        if extra_format_unmarshallers is None:
            extra_format_unmarshallers = {}
        self.extra_format_unmarshallers = extra_format_unmarshallers
        if custom_formatters is None:
            custom_formatters = {}
        self.custom_formatters = custom_formatters

    def unmarshal(self, schema_format: str, value: Any) -> Any:
        format_unmarshaller = self.get_unmarshaller(schema_format)
        if format_unmarshaller is None:
            return value
        try:
            return format_unmarshaller(value)
        except (ValueError, TypeError) as exc:
            raise FormatUnmarshalError(value, schema_format, exc)

    def get_unmarshaller(
        self, schema_format: str
    ) -> Optional[FormatUnmarshaller]:
        if schema_format in self.custom_formatters:
            formatter = self.custom_formatters[schema_format]
            return formatter.format
        if schema_format in self.extra_format_unmarshallers:
            return self.extra_format_unmarshallers[schema_format]
        if schema_format in self.format_unmarshallers:
            return self.format_unmarshallers[schema_format]

        return None

    def __contains__(self, schema_format: str) -> bool:
        format_unmarshallers_dicts: List[Mapping[str, Any]] = [
            self.custom_formatters,
            self.extra_format_unmarshallers,
            self.format_unmarshallers,
        ]
        for content in format_unmarshallers_dicts:
            if schema_format in content:
                return True
        return False


class SchemaUnmarshaller:
    def __init__(
        self,
        schema: Spec,
        schema_validator: SchemaValidator,
        types_unmarshaller: TypesUnmarshaller,
        formats_unmarshaller: FormatsUnmarshaller,
    ):
        self.schema = schema
        self.schema_validator = schema_validator

        self.types_unmarshaller = types_unmarshaller
        self.formats_unmarshaller = formats_unmarshaller

    def __call__(self, value: Any) -> Any:
        warnings.warn(
            "Calling unmarshaller itself is deprecated. "
            "Use unmarshal method instead.",
            DeprecationWarning,
        )
        return self.unmarshal(value)

    def unmarshal(self, value: Any) -> Any:
        self.schema_validator.validate(value)

        # skip unmarshalling for nullable in OpenAPI 3.0
        if value is None and self.schema.getkey("nullable", False):
            return value

        schema_type = self.schema.getkey("type")
        type_unmarshaller = self.get_type_unmarshaller(schema_type)
        typed = type_unmarshaller(value)
        schema_format = self.find_format(value)
        if schema_format is None:
            return typed
        return self.formats_unmarshaller.unmarshal(schema_format, typed)

    def get_type_unmarshaller(
        self,
        schema_type: Optional[Union[Iterable[str], str]],
    ) -> PrimitiveUnmarshaller:
        klass = self.types_unmarshaller.get_unmarshaller(schema_type)
        return klass(
            self.schema,
            self.schema_validator,
            self,
        )

    def evolve(self, schema: Spec) -> "SchemaUnmarshaller":
        cls = self.__class__

        return cls(
            schema,
            self.schema_validator.evolve(schema),
            self.types_unmarshaller,
            self.formats_unmarshaller,
        )

    def find_format(self, value: Any) -> Optional[str]:
        for schema in self.iter_valid_schemas(value):
            if "format" in schema:
                return str(schema.getkey("format"))
        return None

    def iter_valid_schemas(self, value: Any) -> Iterator[Spec]:
        yield self.schema

        one_of_schema = self.schema_validator.get_one_of_schema(value)
        if one_of_schema is not None:
            yield one_of_schema

        yield from self.schema_validator.iter_any_of_schemas(value)
        yield from self.schema_validator.iter_all_of_schemas(value)



"""OpenAPI core contrib falcon middlewares module"""
from typing import Any
from typing import Optional
from typing import Type

from falcon.request import Request
from falcon.response import Response

from openapi_core.contrib.falcon.handlers import FalconOpenAPIErrorsHandler
from openapi_core.contrib.falcon.requests import FalconOpenAPIRequest
from openapi_core.contrib.falcon.responses import FalconOpenAPIResponse
from openapi_core.spec import Spec
from openapi_core.unmarshalling.processors import UnmarshallingProcessor
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType


class FalconOpenAPIMiddleware(UnmarshallingProcessor):
    request_class = FalconOpenAPIRequest
    response_class = FalconOpenAPIResponse
    errors_handler = FalconOpenAPIErrorsHandler()

    def __init__(
        self,
        spec: Spec,
        request_unmarshaller_cls: Optional[RequestUnmarshallerType] = None,
        response_unmarshaller_cls: Optional[ResponseUnmarshallerType] = None,
        request_class: Type[FalconOpenAPIRequest] = FalconOpenAPIRequest,
        response_class: Type[FalconOpenAPIResponse] = FalconOpenAPIResponse,
        errors_handler: Optional[FalconOpenAPIErrorsHandler] = None,
    ):
        super().__init__(
            spec,
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
        )
        self.request_class = request_class or self.request_class
        self.response_class = response_class or self.response_class
        self.errors_handler = errors_handler or self.errors_handler

    @classmethod
    def from_spec(
        cls,
        spec: Spec,
        request_unmarshaller_cls: Optional[RequestUnmarshallerType] = None,
        response_unmarshaller_cls: Optional[ResponseUnmarshallerType] = None,
        request_class: Type[FalconOpenAPIRequest] = FalconOpenAPIRequest,
        response_class: Type[FalconOpenAPIResponse] = FalconOpenAPIResponse,
        errors_handler: Optional[FalconOpenAPIErrorsHandler] = None,
    ) -> "FalconOpenAPIMiddleware":
        return cls(
            spec,
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
            request_class=request_class,
            response_class=response_class,
            errors_handler=errors_handler,
        )

    def process_request(self, req: Request, resp: Response) -> None:  # type: ignore
        openapi_req = self._get_openapi_request(req)
        req.context.openapi = super().process_request(openapi_req)
        if req.context.openapi.errors:
            return self._handle_request_errors(req, resp, req.context.openapi)

    def process_response(  # type: ignore
        self, req: Request, resp: Response, resource: Any, req_succeeded: bool
    ) -> None:
        openapi_req = self._get_openapi_request(req)
        openapi_resp = self._get_openapi_response(resp)
        resp.context.openapi = super().process_response(
            openapi_req, openapi_resp
        )
        if resp.context.openapi.errors:
            return self._handle_response_errors(
                req, resp, resp.context.openapi
            )

    def _handle_request_errors(
        self,
        req: Request,
        resp: Response,
        request_result: RequestUnmarshalResult,
    ) -> None:
        return self.errors_handler.handle(req, resp, request_result.errors)

    def _handle_response_errors(
        self,
        req: Request,
        resp: Response,
        response_result: ResponseUnmarshalResult,
    ) -> None:
        return self.errors_handler.handle(req, resp, response_result.errors)

    def _get_openapi_request(self, request: Request) -> FalconOpenAPIRequest:
        return self.request_class(request)

    def _get_openapi_response(
        self, response: Response
    ) -> FalconOpenAPIResponse:
        return self.response_class(response)



"""OpenAPI core validation response validators module"""
import warnings
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Mapping

from openapi_core.exceptions import OpenAPIError
from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.protocols import WebhookRequest
from openapi_core.spec import Spec
from openapi_core.templating.paths.exceptions import PathError
from openapi_core.templating.responses.exceptions import ResponseFinderError
from openapi_core.validation.decorators import ValidationErrorWrapper
from openapi_core.validation.exceptions import ValidationError
from openapi_core.validation.response.exceptions import DataValidationError
from openapi_core.validation.response.exceptions import HeadersError
from openapi_core.validation.response.exceptions import HeaderValidationError
from openapi_core.validation.response.exceptions import InvalidData
from openapi_core.validation.response.exceptions import InvalidHeader
from openapi_core.validation.response.exceptions import MissingData
from openapi_core.validation.response.exceptions import MissingHeader
from openapi_core.validation.response.exceptions import MissingRequiredHeader
from openapi_core.validation.schemas import (
    oas30_read_schema_validators_factory,
)
from openapi_core.validation.schemas import oas31_schema_validators_factory
from openapi_core.validation.validators import BaseAPICallValidator
from openapi_core.validation.validators import BaseValidator
from openapi_core.validation.validators import BaseWebhookValidator


class BaseResponseValidator(BaseValidator):
    def _iter_errors(
        self,
        status_code: int,
        data: str,
        headers: Mapping[str, Any],
        mimetype: str,
        operation: Spec,
    ) -> Iterator[Exception]:
        try:
            operation_response = self._get_operation_response(
                status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            yield exc
            return

        try:
            self._get_data(data, mimetype, operation_response)
        except DataValidationError as exc:
            yield exc

        try:
            self._get_headers(headers, operation_response)
        except HeadersError as exc:
            yield from exc.context

    def _iter_data_errors(
        self, status_code: int, data: str, mimetype: str, operation: Spec
    ) -> Iterator[Exception]:
        try:
            operation_response = self._get_operation_response(
                status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            yield exc
            return

        try:
            self._get_data(data, mimetype, operation_response)
        except DataValidationError as exc:
            yield exc

    def _iter_headers_errors(
        self, status_code: int, headers: Mapping[str, Any], operation: Spec
    ) -> Iterator[Exception]:
        try:
            operation_response = self._get_operation_response(
                status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            yield exc
            return

        try:
            self._get_headers(headers, operation_response)
        except HeadersError as exc:
            yield from exc.context

    def _get_operation_response(
        self,
        status_code: int,
        operation: Spec,
    ) -> Spec:
        from openapi_core.templating.responses.finders import ResponseFinder

        finder = ResponseFinder(operation / "responses")
        return finder.find(str(status_code))

    @ValidationErrorWrapper(DataValidationError, InvalidData)
    def _get_data(
        self, data: str, mimetype: str, operation_response: Spec
    ) -> Any:
        if "content" not in operation_response:
            return None

        content = operation_response / "content"

        raw_data = self._get_data_value(data)
        return self._get_content_value(raw_data, mimetype, content)

    def _get_data_value(self, data: str) -> Any:
        if not data:
            raise MissingData

        return data

    def _get_headers(
        self, headers: Mapping[str, Any], operation_response: Spec
    ) -> Dict[str, Any]:
        if "headers" not in operation_response:
            return {}

        response_headers = operation_response / "headers"

        errors: List[OpenAPIError] = []
        validated: Dict[str, Any] = {}
        for name, header in list(response_headers.items()):
            # ignore Content-Type header
            if name.lower() == "content-type":
                continue
            try:
                value = self._get_header(headers, name, header)
            except MissingHeader:
                continue
            except ValidationError as exc:
                errors.append(exc)
                continue
            else:
                validated[name] = value

        if errors:
            raise HeadersError(context=iter(errors), headers=validated)

        return validated

    @ValidationErrorWrapper(HeaderValidationError, InvalidHeader, name="name")
    def _get_header(
        self, headers: Mapping[str, Any], name: str, header: Spec
    ) -> Any:
        deprecated = header.getkey("deprecated", False)
        if deprecated:
            warnings.warn(
                f"{name} header is deprecated",
                DeprecationWarning,
            )

        try:
            return self._get_param_or_header_value(header, headers, name=name)
        except KeyError:
            required = header.getkey("required", False)
            if required:
                raise MissingRequiredHeader(name)
            raise MissingHeader(name)


class BaseAPICallResponseValidator(
    BaseResponseValidator, BaseAPICallValidator
):
    def iter_errors(
        self,
        request: Request,
        response: Response,
    ) -> Iterator[Exception]:
        raise NotImplementedError

    def validate(
        self,
        request: Request,
        response: Response,
    ) -> None:
        for err in self.iter_errors(request, response):
            raise err


class BaseWebhookResponseValidator(
    BaseResponseValidator, BaseWebhookValidator
):
    def iter_errors(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> Iterator[Exception]:
        raise NotImplementedError

    def validate(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> None:
        for err in self.iter_errors(request, response):
            raise err


class APICallResponseDataValidator(BaseAPICallResponseValidator):
    def iter_errors(
        self,
        request: Request,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_data_errors(
            response.status_code, response.data, response.mimetype, operation
        )


class APICallResponseHeadersValidator(BaseAPICallResponseValidator):
    def iter_errors(
        self,
        request: Request,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_headers_errors(
            response.status_code, response.headers, operation
        )


class APICallResponseValidator(BaseAPICallResponseValidator):
    def iter_errors(
        self,
        request: Request,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_errors(
            response.status_code,
            response.data,
            response.headers,
            response.mimetype,
            operation,
        )


class WebhookResponseDataValidator(BaseWebhookResponseValidator):
    def iter_errors(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_data_errors(
            response.status_code, response.data, response.mimetype, operation
        )


class WebhookResponseHeadersValidator(BaseWebhookResponseValidator):
    def iter_errors(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_headers_errors(
            response.status_code, response.headers, operation
        )


class WebhookResponseValidator(BaseWebhookResponseValidator):
    def iter_errors(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_errors(
            response.status_code,
            response.data,
            response.headers,
            response.mimetype,
            operation,
        )


class V30ResponseDataValidator(APICallResponseDataValidator):
    schema_validators_factory = oas30_read_schema_validators_factory


class V30ResponseHeadersValidator(APICallResponseHeadersValidator):
    schema_validators_factory = oas30_read_schema_validators_factory


class V30ResponseValidator(APICallResponseValidator):
    schema_validators_factory = oas30_read_schema_validators_factory


class V31ResponseDataValidator(APICallResponseDataValidator):
    schema_validators_factory = oas31_schema_validators_factory


class V31ResponseHeadersValidator(APICallResponseHeadersValidator):
    schema_validators_factory = oas31_schema_validators_factory


class V31ResponseValidator(APICallResponseValidator):
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookResponseDataValidator(WebhookResponseDataValidator):
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookResponseHeadersValidator(WebhookResponseHeadersValidator):
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookResponseValidator(WebhookResponseValidator):
    schema_validators_factory = oas31_schema_validators_factory



import warnings
from typing import Any
from typing import Callable
from typing import List
from typing import Optional

from openapi_core.deserializing.exceptions import DeserializeError
from openapi_core.deserializing.parameters.datatypes import (
    DeserializerCallable,
)
from openapi_core.deserializing.parameters.exceptions import (
    EmptyQueryParameterValue,
)
from openapi_core.schema.parameters import get_aslist
from openapi_core.schema.parameters import get_explode
from openapi_core.spec import Spec


class CallableParameterDeserializer:
    def __init__(
        self,
        param_or_header: Spec,
        style: str,
        deserializer_callable: Optional[DeserializerCallable] = None,
    ):
        self.param_or_header = param_or_header
        self.style = style
        self.deserializer_callable = deserializer_callable

        self.aslist = get_aslist(self.param_or_header)
        self.explode = get_explode(self.param_or_header)

    def deserialize(self, value: Any) -> Any:
        if self.deserializer_callable is None:
            warnings.warn(f"Unsupported {self.style} style")
            return value

        # if "in" not defined then it's a Header
        if "allowEmptyValue" in self.param_or_header:
            warnings.warn(
                "Use of allowEmptyValue property is deprecated",
                DeprecationWarning,
            )
        allow_empty_values = self.param_or_header.getkey(
            "allowEmptyValue", False
        )
        location_name = self.param_or_header.getkey("in", "header")
        if location_name == "query" and value == "" and not allow_empty_values:
            name = self.param_or_header["name"]
            raise EmptyQueryParameterValue(name)

        if not self.aslist or self.explode:
            return value
        try:
            return self.deserializer_callable(value)
        except (ValueError, TypeError, AttributeError):
            raise DeserializeError(location_name, self.style, value)



"""OpenAPI core contrib flask decorators module"""
from functools import wraps
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type

from flask.globals import request
from flask.helpers import make_response
from flask.wrappers import Request
from flask.wrappers import Response

from openapi_core.contrib.flask.handlers import FlaskOpenAPIErrorsHandler
from openapi_core.contrib.flask.providers import FlaskRequestProvider
from openapi_core.contrib.flask.requests import FlaskOpenAPIRequest
from openapi_core.contrib.flask.responses import FlaskOpenAPIResponse
from openapi_core.spec import Spec
from openapi_core.unmarshalling.processors import UnmarshallingProcessor
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType


class FlaskOpenAPIViewDecorator(UnmarshallingProcessor):
    def __init__(
        self,
        spec: Spec,
        request_unmarshaller_cls: Optional[RequestUnmarshallerType] = None,
        response_unmarshaller_cls: Optional[ResponseUnmarshallerType] = None,
        request_class: Type[FlaskOpenAPIRequest] = FlaskOpenAPIRequest,
        response_class: Type[FlaskOpenAPIResponse] = FlaskOpenAPIResponse,
        request_provider: Type[FlaskRequestProvider] = FlaskRequestProvider,
        openapi_errors_handler: Type[
            FlaskOpenAPIErrorsHandler
        ] = FlaskOpenAPIErrorsHandler,
    ):
        super().__init__(
            spec,
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
        )
        self.request_class = request_class
        self.response_class = response_class
        self.request_provider = request_provider
        self.openapi_errors_handler = openapi_errors_handler

    def __call__(self, view: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(view)
        def decorated(*args: Any, **kwargs: Any) -> Response:
            request = self._get_request()
            openapi_request = self._get_openapi_request(request)
            request_result = self.process_request(openapi_request)
            if request_result.errors:
                return self._handle_request_errors(request_result)
            response = self._handle_request_view(
                request_result, view, *args, **kwargs
            )
            openapi_response = self._get_openapi_response(response)
            response_result = self.process_response(
                openapi_request, openapi_response
            )
            if response_result.errors:
                return self._handle_response_errors(response_result)
            return response

        return decorated

    def _handle_request_view(
        self,
        request_result: RequestUnmarshalResult,
        view: Callable[[Any], Response],
        *args: Any,
        **kwargs: Any
    ) -> Response:
        request = self._get_request()
        request.openapi = request_result  # type: ignore
        rv = view(*args, **kwargs)
        return make_response(rv)

    def _handle_request_errors(
        self, request_result: RequestUnmarshalResult
    ) -> Response:
        return self.openapi_errors_handler.handle(request_result.errors)

    def _handle_response_errors(
        self, response_result: ResponseUnmarshalResult
    ) -> Response:
        return self.openapi_errors_handler.handle(response_result.errors)

    def _get_request(self) -> Request:
        return request

    def _get_openapi_request(self, request: Request) -> FlaskOpenAPIRequest:
        return self.request_class(request)

    def _get_openapi_response(
        self, response: Response
    ) -> FlaskOpenAPIResponse:
        return self.response_class(response)

    @classmethod
    def from_spec(
        cls,
        spec: Spec,
        request_unmarshaller_cls: Optional[RequestUnmarshallerType] = None,
        response_unmarshaller_cls: Optional[ResponseUnmarshallerType] = None,
        request_class: Type[FlaskOpenAPIRequest] = FlaskOpenAPIRequest,
        response_class: Type[FlaskOpenAPIResponse] = FlaskOpenAPIResponse,
        request_provider: Type[FlaskRequestProvider] = FlaskRequestProvider,
        openapi_errors_handler: Type[
            FlaskOpenAPIErrorsHandler
        ] = FlaskOpenAPIErrorsHandler,
    ) -> "FlaskOpenAPIViewDecorator":
        return cls(
            spec,
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
            request_class=request_class,
            response_class=response_class,
            request_provider=request_provider,
            openapi_errors_handler=openapi_errors_handler,
        )



"""OpenAPI spec validator validation proxies module."""
import warnings
from typing import TYPE_CHECKING
from typing import Any
from typing import Iterator
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Type

from openapi_core.exceptions import SpecError
from openapi_core.protocols import Request
from openapi_core.spec import Spec
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult

if TYPE_CHECKING:
    from openapi_core.unmarshalling.request.unmarshallers import (
        APICallRequestUnmarshaller,
    )


class SpecRequestValidatorProxy:
    def __init__(
        self,
        unmarshaller_cls: Type["APICallRequestUnmarshaller"],
        deprecated: str = "RequestValidator",
        use: Optional[str] = None,
        **unmarshaller_kwargs: Any,
    ):
        self.unmarshaller_cls = unmarshaller_cls
        self.unmarshaller_kwargs = unmarshaller_kwargs

        self.deprecated = deprecated
        self.use = use or self.unmarshaller_cls.__name__

    def validate(
        self,
        spec: Spec,
        request: Request,
        base_url: Optional[str] = None,
    ) -> "RequestUnmarshalResult":
        warnings.warn(
            f"{self.deprecated} is deprecated. Use {self.use} instead.",
            DeprecationWarning,
        )
        unmarshaller = self.unmarshaller_cls(
            spec, base_url=base_url, **self.unmarshaller_kwargs
        )
        return unmarshaller.unmarshal(request)

    def is_valid(
        self,
        spec: Spec,
        request: Request,
        base_url: Optional[str] = None,
    ) -> bool:
        unmarshaller = self.unmarshaller_cls(
            spec, base_url=base_url, **self.unmarshaller_kwargs
        )
        error = next(unmarshaller.iter_errors(request), None)
        return error is None

    def iter_errors(
        self,
        spec: Spec,
        request: Request,
        base_url: Optional[str] = None,
    ) -> Iterator[Exception]:
        unmarshaller = self.unmarshaller_cls(
            spec, base_url=base_url, **self.unmarshaller_kwargs
        )
        yield from unmarshaller.iter_errors(request)


class DetectSpecRequestValidatorProxy:
    def __init__(
        self, choices: Mapping[Tuple[str, str], SpecRequestValidatorProxy]
    ):
        self.choices = choices

    def detect(self, spec: Spec) -> SpecRequestValidatorProxy:
        for (key, value), validator in self.choices.items():
            if key in spec and spec[key].startswith(value):
                return validator
        raise SpecError("Spec schema version not detected")

    def validate(
        self,
        spec: Spec,
        request: Request,
        base_url: Optional[str] = None,
    ) -> "RequestUnmarshalResult":
        validator = self.detect(spec)
        return validator.validate(spec, request, base_url=base_url)

    def is_valid(
        self,
        spec: Spec,
        request: Request,
        base_url: Optional[str] = None,
    ) -> bool:
        validator = self.detect(spec)
        error = next(
            validator.iter_errors(spec, request, base_url=base_url), None
        )
        return error is None

    def iter_errors(
        self,
        spec: Spec,
        request: Request,
        base_url: Optional[str] = None,
    ) -> Iterator[Exception]:
        validator = self.detect(spec)
        yield from validator.iter_errors(spec, request, base_url=base_url)



import re
from functools import partial
from typing import Dict

from openapi_core.deserializing.parameters.datatypes import (
    DeserializerCallable,
)
from openapi_core.deserializing.parameters.deserializers import (
    CallableParameterDeserializer,
)
from openapi_core.deserializing.parameters.util import split
from openapi_core.schema.parameters import get_style
from openapi_core.spec import Spec


class ParameterDeserializersFactory:
    PARAMETER_STYLE_DESERIALIZERS: Dict[str, DeserializerCallable] = {
        "form": partial(split, separator=","),
        "simple": partial(split, separator=","),
        "spaceDelimited": partial(split, separator=" "),
        "pipeDelimited": partial(split, separator="|"),
        "deepObject": partial(re.split, pattern=r"\[|\]"),
    }

    def create(self, param_or_header: Spec) -> CallableParameterDeserializer:
        style = get_style(param_or_header)

        deserialize_callable = self.PARAMETER_STYLE_DESERIALIZERS.get(style)
        return CallableParameterDeserializer(
            param_or_header, style, deserialize_callable
        )



"""OpenAPI core validation validators module"""
import sys
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Tuple
from urllib.parse import urljoin

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from backports.cached_property import cached_property
from openapi_core.casting.schemas import schema_casters_factory
from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.deserializing.media_types import (
    media_type_deserializers_factory,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.parameters import (
    parameter_deserializers_factory,
)
from openapi_core.deserializing.parameters.factories import (
    ParameterDeserializersFactory,
)
from openapi_core.protocols import Request
from openapi_core.protocols import WebhookRequest
from openapi_core.schema.parameters import get_value
from openapi_core.spec import Spec
from openapi_core.templating.media_types.datatypes import MediaType
from openapi_core.templating.paths.datatypes import PathOperationServer
from openapi_core.templating.paths.finders import APICallPathFinder
from openapi_core.templating.paths.finders import BasePathFinder
from openapi_core.templating.paths.finders import WebhookPathFinder
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory


class BaseValidator:
    schema_validators_factory: SchemaValidatorsFactory = NotImplemented

    def __init__(
        self,
        spec: Spec,
        base_url: Optional[str] = None,
        schema_casters_factory: SchemaCastersFactory = schema_casters_factory,
        parameter_deserializers_factory: ParameterDeserializersFactory = parameter_deserializers_factory,
        media_type_deserializers_factory: MediaTypeDeserializersFactory = media_type_deserializers_factory,
        schema_validators_factory: Optional[SchemaValidatorsFactory] = None,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
    ):
        self.spec = spec
        self.base_url = base_url

        self.schema_casters_factory = schema_casters_factory
        self.parameter_deserializers_factory = parameter_deserializers_factory
        self.media_type_deserializers_factory = (
            media_type_deserializers_factory
        )
        self.schema_validators_factory = (
            schema_validators_factory or self.schema_validators_factory
        )
        if self.schema_validators_factory is NotImplemented:
            raise NotImplementedError(
                "schema_validators_factory is not assigned"
            )
        self.format_validators = format_validators
        self.extra_format_validators = extra_format_validators
        self.extra_media_type_deserializers = extra_media_type_deserializers

    def _get_media_type(self, content: Spec, mimetype: str) -> MediaType:
        from openapi_core.templating.media_types.finders import MediaTypeFinder

        finder = MediaTypeFinder(content)
        return finder.find(mimetype)

    def _deserialise_data(self, mimetype: str, value: Any) -> Any:
        deserializer = self.media_type_deserializers_factory.create(
            mimetype,
            extra_media_type_deserializers=self.extra_media_type_deserializers,
        )
        return deserializer.deserialize(value)

    def _deserialise_parameter(self, param: Spec, value: Any) -> Any:
        deserializer = self.parameter_deserializers_factory.create(param)
        return deserializer.deserialize(value)

    def _cast(self, schema: Spec, value: Any) -> Any:
        caster = self.schema_casters_factory.create(schema)
        return caster(value)

    def _validate_schema(self, schema: Spec, value: Any) -> None:
        validator = self.schema_validators_factory.create(
            schema,
            format_validators=self.format_validators,
            extra_format_validators=self.extra_format_validators,
        )
        validator.validate(value)

    def _get_param_or_header_value(
        self,
        param_or_header: Spec,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Any:
        casted, schema = self._get_param_or_header_value_and_schema(
            param_or_header, location, name
        )
        if schema is None:
            return casted
        self._validate_schema(schema, casted)
        return casted

    def _get_content_value(
        self, raw: Any, mimetype: str, content: Spec
    ) -> Any:
        casted, schema = self._get_content_value_and_schema(
            raw, mimetype, content
        )
        if schema is None:
            return casted
        self._validate_schema(schema, casted)
        return casted

    def _get_param_or_header_value_and_schema(
        self,
        param_or_header: Spec,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Tuple[Any, Spec]:
        try:
            raw_value = get_value(param_or_header, location, name=name)
        except KeyError:
            if "schema" not in param_or_header:
                raise
            schema = param_or_header / "schema"
            if "default" not in schema:
                raise
            casted = schema["default"]
        else:
            # Simple scenario
            if "content" not in param_or_header:
                deserialised = self._deserialise_parameter(
                    param_or_header, raw_value
                )
                schema = param_or_header / "schema"
            # Complex scenario
            else:
                content = param_or_header / "content"
                mimetype, media_type = next(content.items())
                deserialised = self._deserialise_data(mimetype, raw_value)
                schema = media_type / "schema"
            casted = self._cast(schema, deserialised)
        return casted, schema

    def _get_content_value_and_schema(
        self, raw: Any, mimetype: str, content: Spec
    ) -> Tuple[Any, Optional[Spec]]:
        media_type, mimetype = self._get_media_type(content, mimetype)
        deserialised = self._deserialise_data(mimetype, raw)
        casted = self._cast(media_type, deserialised)

        if "schema" not in media_type:
            return casted, None

        schema = media_type / "schema"
        return casted, schema


class BaseAPICallValidator(BaseValidator):
    @cached_property
    def path_finder(self) -> BasePathFinder:
        return APICallPathFinder(self.spec, base_url=self.base_url)

    def _find_path(self, request: Request) -> PathOperationServer:
        path_pattern = getattr(request, "path_pattern", None) or request.path
        full_url = urljoin(request.host_url, path_pattern)
        return self.path_finder.find(request.method, full_url)


class BaseWebhookValidator(BaseValidator):
    @cached_property
    def path_finder(self) -> BasePathFinder:
        return WebhookPathFinder(self.spec, base_url=self.base_url)

    def _find_path(self, request: WebhookRequest) -> PathOperationServer:
        return self.path_finder.find(request.method, request.name)



import warnings
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type


class Formatter:
    def validate(self, value: Any) -> bool:
        return True

    def format(self, value: Any) -> Any:
        return value

    def __getattribute__(self, name: str) -> Any:
        if name == "unmarshal":
            warnings.warn(
                "Unmarshal method is deprecated. " "Use format instead.",
                DeprecationWarning,
            )
            return super().__getattribute__("format")
        if name == "format":
            try:
                attr = super().__getattribute__("unmarshal")
            except AttributeError:
                return super().__getattribute__("format")
            else:
                warnings.warn(
                    "Unmarshal method is deprecated. "
                    "Rename unmarshal method to format instead.",
                    DeprecationWarning,
                )
                return attr
        return super().__getattribute__(name)

    @classmethod
    def from_callables(
        cls,
        validate_callable: Optional[Callable[[Any], Any]] = None,
        format_callable: Optional[Callable[[Any], Any]] = None,
        unmarshal: Optional[Callable[[Any], Any]] = None,
    ) -> "Formatter":
        attrs = {}
        if validate_callable is not None:
            attrs["validate"] = staticmethod(validate_callable)
        if format_callable is not None:
            attrs["format"] = staticmethod(format_callable)
        if unmarshal is not None:
            warnings.warn(
                "Unmarshal parameter is deprecated. "
                "Use format_callable instead.",
                DeprecationWarning,
            )
            attrs["format"] = staticmethod(unmarshal)

        klass: Type[Formatter] = type("Formatter", (cls,), attrs)
        return klass()



from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

from openapi_core.validation.schemas.formatters import Formatter

FormatValidator = Callable[[Any], bool]

CustomFormattersDict = Dict[str, Formatter]
FormattersDict = Dict[Optional[str], Formatter]
FormatValidatorsDict = Dict[str, FormatValidator]



from typing import Tuple

from openapi_core.spec.paths import Spec
from openapi_core.templating.paths.datatypes import Path


def template_path_len(template_path: Path) -> int:
    return len(template_path[1].variables)



from openapi_core.deserializing.parameters.factories import (
    ParameterDeserializersFactory,
)

__all__ = ["parameter_deserializers_factory"]

parameter_deserializers_factory = ParameterDeserializersFactory()



"""OpenAPI core shortcuts module"""
import warnings
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from openapi_core.exceptions import SpecError
from openapi_core.finders import SpecClasses
from openapi_core.finders import SpecFinder
from openapi_core.finders import SpecVersion
from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.protocols import WebhookRequest
from openapi_core.spec import Spec
from openapi_core.unmarshalling.request import V30RequestUnmarshaller
from openapi_core.unmarshalling.request import V31RequestUnmarshaller
from openapi_core.unmarshalling.request import V31WebhookRequestUnmarshaller
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.protocols import RequestUnmarshaller
from openapi_core.unmarshalling.request.protocols import (
    WebhookRequestUnmarshaller,
)
from openapi_core.unmarshalling.request.proxies import (
    SpecRequestValidatorProxy,
)
from openapi_core.unmarshalling.request.types import AnyRequestUnmarshallerType
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.request.types import (
    WebhookRequestUnmarshallerType,
)
from openapi_core.unmarshalling.response import V30ResponseUnmarshaller
from openapi_core.unmarshalling.response import V31ResponseUnmarshaller
from openapi_core.unmarshalling.response import V31WebhookResponseUnmarshaller
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.response.protocols import ResponseUnmarshaller
from openapi_core.unmarshalling.response.protocols import (
    WebhookResponseUnmarshaller,
)
from openapi_core.unmarshalling.response.proxies import (
    SpecResponseValidatorProxy,
)
from openapi_core.unmarshalling.response.types import (
    AnyResponseUnmarshallerType,
)
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType
from openapi_core.unmarshalling.response.types import (
    WebhookResponseUnmarshallerType,
)
from openapi_core.validation.request import V30RequestValidator
from openapi_core.validation.request import V31RequestValidator
from openapi_core.validation.request import V31WebhookRequestValidator
from openapi_core.validation.request.protocols import RequestValidator
from openapi_core.validation.request.protocols import WebhookRequestValidator
from openapi_core.validation.request.types import AnyRequestValidatorType
from openapi_core.validation.request.types import RequestValidatorType
from openapi_core.validation.request.types import WebhookRequestValidatorType
from openapi_core.validation.response import V30ResponseValidator
from openapi_core.validation.response import V31ResponseValidator
from openapi_core.validation.response import V31WebhookResponseValidator
from openapi_core.validation.response.protocols import ResponseValidator
from openapi_core.validation.response.protocols import WebhookResponseValidator
from openapi_core.validation.response.types import AnyResponseValidatorType
from openapi_core.validation.response.types import ResponseValidatorType
from openapi_core.validation.response.types import WebhookResponseValidatorType

AnyRequest = Union[Request, WebhookRequest]

SPECS: Dict[SpecVersion, SpecClasses] = {
    SpecVersion("openapi", "3.0"): SpecClasses(
        V30RequestValidator,
        V30ResponseValidator,
        None,
        None,
        V30RequestUnmarshaller,
        V30ResponseUnmarshaller,
        None,
        None,
    ),
    SpecVersion("openapi", "3.1"): SpecClasses(
        V31RequestValidator,
        V31ResponseValidator,
        V31WebhookRequestValidator,
        V31WebhookResponseValidator,
        V31RequestUnmarshaller,
        V31ResponseUnmarshaller,
        V31WebhookRequestUnmarshaller,
        V31WebhookResponseUnmarshaller,
    ),
}


def get_classes(spec: Spec) -> SpecClasses:
    return SpecFinder(SPECS).get_classes(spec)


def unmarshal_apicall_request(
    request: Request,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[RequestUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> RequestUnmarshalResult:
    if not isinstance(request, Request):
        raise TypeError("'request' argument is not type of Request")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if cls is None:
        classes = get_classes(spec)
        cls = classes.request_unmarshaller_cls
    if not issubclass(cls, RequestUnmarshaller):
        raise TypeError("'cls' argument is not type of RequestUnmarshaller")
    v = cls(spec, base_url=base_url, **unmarshaller_kwargs)
    result = v.unmarshal(request)
    result.raise_for_errors()
    return result


def unmarshal_webhook_request(
    request: WebhookRequest,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[WebhookRequestUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> RequestUnmarshalResult:
    if not isinstance(request, WebhookRequest):
        raise TypeError("'request' argument is not type of WebhookRequest")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if cls is None:
        classes = get_classes(spec)
        cls = classes.webhook_request_unmarshaller_cls
        if cls is None:
            raise SpecError("Unmarshaller class not found")
    if not issubclass(cls, WebhookRequestUnmarshaller):
        raise TypeError(
            "'cls' argument is not type of WebhookRequestUnmarshaller"
        )
    v = cls(spec, base_url=base_url, **unmarshaller_kwargs)
    result = v.unmarshal(request)
    result.raise_for_errors()
    return result


def unmarshal_request(
    request: AnyRequest,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[AnyRequestUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> RequestUnmarshalResult:
    if not isinstance(request, (Request, WebhookRequest)):
        raise TypeError("'request' argument is not type of (Webhook)Request")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if isinstance(request, WebhookRequest):
        if cls is None or issubclass(cls, WebhookRequestUnmarshaller):
            return unmarshal_webhook_request(
                request,
                spec,
                base_url=base_url,
                cls=cls,
                **unmarshaller_kwargs,
            )
        else:
            raise TypeError(
                "'cls' argument is not type of WebhookRequestUnmarshaller"
            )
    else:
        if cls is None or issubclass(cls, RequestUnmarshaller):
            return unmarshal_apicall_request(
                request,
                spec,
                base_url=base_url,
                cls=cls,
                **unmarshaller_kwargs,
            )
        else:
            raise TypeError(
                "'cls' argument is not type of RequestUnmarshaller"
            )


def unmarshal_apicall_response(
    request: Request,
    response: Response,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[ResponseUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> ResponseUnmarshalResult:
    if not isinstance(request, Request):
        raise TypeError("'request' argument is not type of Request")
    if not isinstance(response, Response):
        raise TypeError("'response' argument is not type of Response")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if cls is None:
        classes = get_classes(spec)
        cls = classes.response_unmarshaller_cls
    if not issubclass(cls, ResponseUnmarshaller):
        raise TypeError("'cls' argument is not type of ResponseUnmarshaller")
    v = cls(spec, base_url=base_url, **unmarshaller_kwargs)
    result = v.unmarshal(request, response)
    result.raise_for_errors()
    return result


def unmarshal_webhook_response(
    request: WebhookRequest,
    response: Response,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[WebhookResponseUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> ResponseUnmarshalResult:
    if not isinstance(request, WebhookRequest):
        raise TypeError("'request' argument is not type of WebhookRequest")
    if not isinstance(response, Response):
        raise TypeError("'response' argument is not type of Response")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if cls is None:
        classes = get_classes(spec)
        cls = classes.webhook_response_unmarshaller_cls
        if cls is None:
            raise SpecError("Unmarshaller class not found")
    if not issubclass(cls, WebhookResponseUnmarshaller):
        raise TypeError(
            "'cls' argument is not type of WebhookResponseUnmarshaller"
        )
    v = cls(spec, base_url=base_url, **unmarshaller_kwargs)
    result = v.unmarshal(request, response)
    result.raise_for_errors()
    return result


def unmarshal_response(
    request: AnyRequest,
    response: Response,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[AnyResponseUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> ResponseUnmarshalResult:
    if not isinstance(request, (Request, WebhookRequest)):
        raise TypeError("'request' argument is not type of (Webhook)Request")
    if not isinstance(response, Response):
        raise TypeError("'response' argument is not type of Response")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if isinstance(request, WebhookRequest):
        if cls is None or issubclass(cls, WebhookResponseUnmarshaller):
            return unmarshal_webhook_response(
                request,
                response,
                spec,
                base_url=base_url,
                cls=cls,
                **unmarshaller_kwargs,
            )
        else:
            raise TypeError(
                "'cls' argument is not type of WebhookResponseUnmarshaller"
            )
    else:
        if cls is None or issubclass(cls, ResponseUnmarshaller):
            return unmarshal_apicall_response(
                request,
                response,
                spec,
                base_url=base_url,
                cls=cls,
                **unmarshaller_kwargs,
            )
        else:
            raise TypeError(
                "'cls' argument is not type of ResponseUnmarshaller"
            )


def validate_request(
    request: AnyRequest,
    spec: Spec,
    base_url: Optional[str] = None,
    validator: Optional[SpecRequestValidatorProxy] = None,
    cls: Optional[AnyRequestValidatorType] = None,
    **validator_kwargs: Any,
) -> Optional[RequestUnmarshalResult]:
    if isinstance(spec, (Request, WebhookRequest)) and isinstance(
        request, Spec
    ):
        warnings.warn(
            "spec parameter as a first argument is deprecated. "
            "Move it to second argument instead.",
            DeprecationWarning,
        )
        request, spec = spec, request

    if not isinstance(request, (Request, WebhookRequest)):
        raise TypeError("'request' argument is not type of (Webhook)Request")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")

    if validator is not None and isinstance(request, Request):
        warnings.warn(
            "validator parameter is deprecated. Use cls instead.",
            DeprecationWarning,
        )
        result = validator.validate(spec, request, base_url=base_url)
        result.raise_for_errors()
        return result

    # redirect to unmarshaller for backward compatibility
    if cls is None or issubclass(
        cls, (RequestUnmarshaller, WebhookRequestUnmarshaller)
    ):
        warnings.warn(
            "validate_request is deprecated for unmarshalling data "
            "and it will not return any result in the future. "
            "Use unmarshal_request function instead.",
            DeprecationWarning,
        )
        return unmarshal_request(
            request,
            spec=spec,
            base_url=base_url,
            cls=cls,
            **validator_kwargs,
        )
    if isinstance(request, WebhookRequest):
        if cls is None or issubclass(cls, WebhookRequestValidator):
            validate_webhook_request(
                request,
                spec,
                base_url=base_url,
                cls=cls,
                **validator_kwargs,
            )
            return None
        else:
            raise TypeError(
                "'cls' argument is not type of WebhookRequestValidator"
            )
    else:
        if cls is None or issubclass(cls, RequestValidator):
            validate_apicall_request(
                request,
                spec,
                base_url=base_url,
                cls=cls,
                **validator_kwargs,
            )
            return None
        else:
            raise TypeError("'cls' argument is not type of RequestValidator")


def validate_response(
    request: Union[Request, WebhookRequest, Spec],
    response: Union[Response, Request, WebhookRequest],
    spec: Union[Spec, Response],
    base_url: Optional[str] = None,
    validator: Optional[SpecResponseValidatorProxy] = None,
    cls: Optional[AnyResponseValidatorType] = None,
    **validator_kwargs: Any,
) -> Optional[ResponseUnmarshalResult]:
    if (
        isinstance(request, Spec)
        and isinstance(response, (Request, WebhookRequest))
        and isinstance(spec, Response)
    ):
        warnings.warn(
            "spec parameter as a first argument is deprecated. "
            "Move it to third argument instead.",
            DeprecationWarning,
        )
        args = request, response, spec
        spec, request, response = args

    if not isinstance(request, (Request, WebhookRequest)):
        raise TypeError("'request' argument is not type of (Webhook)Request")
    if not isinstance(response, Response):
        raise TypeError("'response' argument is not type of Response")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")

    if validator is not None and isinstance(request, Request):
        warnings.warn(
            "validator parameter is deprecated. Use cls instead.",
            DeprecationWarning,
        )
        result = validator.validate(spec, request, response, base_url=base_url)
        result.raise_for_errors()
        return result

    # redirect to unmarshaller for backward compatibility
    if cls is None or issubclass(
        cls, (ResponseUnmarshaller, WebhookResponseUnmarshaller)
    ):
        warnings.warn(
            "validate_response is deprecated for unmarshalling data "
            "and it will not return any result in the future. "
            "Use unmarshal_response function instead.",
            DeprecationWarning,
        )
        return unmarshal_response(
            request,
            response,
            spec=spec,
            base_url=base_url,
            cls=cls,
            **validator_kwargs,
        )
    if isinstance(request, WebhookRequest):
        if cls is None or issubclass(cls, WebhookResponseValidator):
            validate_webhook_response(
                request,
                response,
                spec,
                base_url=base_url,
                cls=cls,
                **validator_kwargs,
            )
            return None
        else:
            raise TypeError(
                "'cls' argument is not type of WebhookResponseValidator"
            )
    else:
        if cls is None or issubclass(cls, ResponseValidator):
            validate_apicall_response(
                request,
                response,
                spec,
                base_url=base_url,
                cls=cls,
                **validator_kwargs,
            )
            return None
        else:
            raise TypeError("'cls' argument is not type of ResponseValidator")


def validate_apicall_request(
    request: Request,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[RequestValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    if not isinstance(request, Request):
        raise TypeError("'request' argument is not type of Request")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if cls is None:
        classes = get_classes(spec)
        cls = classes.request_validator_cls
    if not issubclass(cls, RequestValidator):
        raise TypeError("'cls' argument is not type of RequestValidator")
    v = cls(spec, base_url=base_url, **validator_kwargs)
    return v.validate(request)


def validate_webhook_request(
    request: WebhookRequest,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[WebhookRequestValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    if not isinstance(request, WebhookRequest):
        raise TypeError("'request' argument is not type of WebhookRequest")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if cls is None:
        classes = get_classes(spec)
        cls = classes.webhook_request_validator_cls
        if cls is None:
            raise SpecError("Validator class not found")
    if not issubclass(cls, WebhookRequestValidator):
        raise TypeError(
            "'cls' argument is not type of WebhookRequestValidator"
        )
    v = cls(spec, base_url=base_url, **validator_kwargs)
    return v.validate(request)


def validate_apicall_response(
    request: Request,
    response: Response,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[ResponseValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    if not isinstance(request, Request):
        raise TypeError("'request' argument is not type of Request")
    if not isinstance(response, Response):
        raise TypeError("'response' argument is not type of Response")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if cls is None:
        classes = get_classes(spec)
        cls = classes.response_validator_cls
    if not issubclass(cls, ResponseValidator):
        raise TypeError("'cls' argument is not type of ResponseValidator")
    v = cls(spec, base_url=base_url, **validator_kwargs)
    return v.validate(request, response)


def validate_webhook_response(
    request: WebhookRequest,
    response: Response,
    spec: Spec,
    base_url: Optional[str] = None,
    cls: Optional[WebhookResponseValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    if not isinstance(request, WebhookRequest):
        raise TypeError("'request' argument is not type of WebhookRequest")
    if not isinstance(response, Response):
        raise TypeError("'response' argument is not type of Response")
    if not isinstance(spec, Spec):
        raise TypeError("'spec' argument is not type of Spec")
    if cls is None:
        classes = get_classes(spec)
        cls = classes.webhook_response_validator_cls
        if cls is None:
            raise SpecError("Validator class not found")
    if not issubclass(cls, WebhookResponseValidator):
        raise TypeError(
            "'cls' argument is not type of WebhookResponseValidator"
        )
    v = cls(spec, base_url=base_url, **validator_kwargs)
    return v.validate(request, response)



"""OpenAPI core unmarshalling response module"""
from openapi_core.unmarshalling.response.proxies import (
    DetectResponseValidatorProxy,
)
from openapi_core.unmarshalling.response.proxies import (
    SpecResponseValidatorProxy,
)
from openapi_core.unmarshalling.response.unmarshallers import (
    APICallResponseUnmarshaller,
)
from openapi_core.unmarshalling.response.unmarshallers import ResponseValidator
from openapi_core.unmarshalling.response.unmarshallers import (
    V30ResponseUnmarshaller,
)
from openapi_core.unmarshalling.response.unmarshallers import (
    V31ResponseUnmarshaller,
)
from openapi_core.unmarshalling.response.unmarshallers import (
    V31WebhookResponseUnmarshaller,
)
from openapi_core.unmarshalling.schemas import (
    oas30_read_schema_unmarshallers_factory,
)
from openapi_core.unmarshalling.schemas import (
    oas31_schema_unmarshallers_factory,
)

__all__ = [
    "V30ResponseUnmarshaller",
    "V31ResponseUnmarshaller",
    "V31WebhookResponseUnmarshaller",
    "ResponseValidator",
    "openapi_v30_response_validator",
    "openapi_v31_response_validator",
    "openapi_v3_response_validator",
    "openapi_response_validator",
]

# alias to the latest v3 version
V3ResponseUnmarshaller = V31ResponseUnmarshaller
V3WebhookResponseUnmarshaller = V31WebhookResponseUnmarshaller

# spec validators
openapi_v30_response_validator = SpecResponseValidatorProxy(
    APICallResponseUnmarshaller,
    schema_unmarshallers_factory=oas30_read_schema_unmarshallers_factory,
    deprecated="openapi_v30_response_validator",
    use="V30ResponseUnmarshaller",
)

openapi_v31_response_validator = SpecResponseValidatorProxy(
    APICallResponseUnmarshaller,
    schema_unmarshallers_factory=oas31_schema_unmarshallers_factory,
    deprecated="openapi_v31_response_validator",
    use="V31ResponseUnmarshaller",
)

# spec validators alias to the latest v3 version
openapi_v3_response_validator = openapi_v31_response_validator

# detect version spec
openapi_response_validator = DetectResponseValidatorProxy(
    {
        ("openapi", "3.0"): openapi_v30_response_validator,
        ("openapi", "3.1"): openapi_v31_response_validator,
    },
)



from typing import Mapping
from typing import NamedTuple
from typing import Optional
from typing import Type

from openapi_core.exceptions import SpecError
from openapi_core.spec import Spec
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.request.types import (
    WebhookRequestUnmarshallerType,
)
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType
from openapi_core.unmarshalling.response.types import (
    WebhookResponseUnmarshallerType,
)
from openapi_core.validation.request.types import RequestValidatorType
from openapi_core.validation.request.types import WebhookRequestValidatorType
from openapi_core.validation.response.types import ResponseValidatorType
from openapi_core.validation.response.types import WebhookResponseValidatorType
from openapi_core.validation.validators import BaseValidator


class SpecVersion(NamedTuple):
    name: str
    version: str


class SpecClasses(NamedTuple):
    request_validator_cls: RequestValidatorType
    response_validator_cls: ResponseValidatorType
    webhook_request_validator_cls: Optional[WebhookRequestValidatorType]
    webhook_response_validator_cls: Optional[WebhookResponseValidatorType]
    request_unmarshaller_cls: RequestUnmarshallerType
    response_unmarshaller_cls: ResponseUnmarshallerType
    webhook_request_unmarshaller_cls: Optional[WebhookRequestUnmarshallerType]
    webhook_response_unmarshaller_cls: Optional[
        WebhookResponseUnmarshallerType
    ]


class SpecFinder:
    def __init__(self, specs: Mapping[SpecVersion, SpecClasses]) -> None:
        self.specs = specs

    def get_classes(self, spec: Spec) -> SpecClasses:
        for v, classes in self.specs.items():
            if v.name in spec and spec[v.name].startswith(v.version):
                return classes
        raise SpecError("Spec schema version not detected")



"""OpenAPI core spec shortcuts module"""
import warnings
from typing import Any
from typing import Dict
from typing import Hashable
from typing import Mapping
from typing import Optional

from jsonschema_spec import default_handlers
from openapi_spec_validator.validation import openapi_spec_validator_proxy
from openapi_spec_validator.validation.protocols import SupportsValidation

from openapi_core.spec.paths import Spec


def create_spec(
    spec_dict: Mapping[Hashable, Any],
    spec_url: str = "",
    handlers: Dict[str, Any] = default_handlers,
    validate_spec: bool = True,
) -> Spec:
    warnings.warn(
        "create_spec function is deprecated. Use Spec.from_dict instead.",
        DeprecationWarning,
    )

    validator: Optional[SupportsValidation] = None
    if validate_spec:
        validator = openapi_spec_validator_proxy

    return Spec.from_dict(
        spec_dict,
        spec_url=spec_url,
        ref_resolver_handlers=handlers,
        validator=validator,
    )



"""OpenAPI X-Model extension factories module"""
from dataclasses import make_dataclass
from pydoc import ErrorDuringImport
from pydoc import locate
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Type

from openapi_core.extensions.models.types import Field
from openapi_core.spec import Spec


class DictFactory:
    base_class = dict

    def create(
        self, schema: Spec, fields: Iterable[Field]
    ) -> Type[Dict[Any, Any]]:
        return self.base_class


class ModelFactory(DictFactory):
    def create(
        self,
        schema: Spec,
        fields: Iterable[Field],
    ) -> Type[Any]:
        name = schema.getkey("x-model")
        if name is None:
            return super().create(schema, fields)

        return make_dataclass(name, fields, frozen=True)


class ModelPathFactory(ModelFactory):
    def create(
        self,
        schema: Spec,
        fields: Iterable[Field],
    ) -> Any:
        model_class_path = schema.getkey("x-model-path")
        if model_class_path is None:
            return super().create(schema, fields)

        return locate(model_class_path)



import warnings
from typing import Any
from typing import Optional

from openapi_core.deserializing.media_types.datatypes import (
    DeserializerCallable,
)
from openapi_core.deserializing.media_types.exceptions import (
    MediaTypeDeserializeError,
)


class CallableMediaTypeDeserializer:
    def __init__(
        self,
        mimetype: str,
        deserializer_callable: Optional[DeserializerCallable] = None,
    ):
        self.mimetype = mimetype
        self.deserializer_callable = deserializer_callable

    def deserialize(self, value: Any) -> Any:
        if self.deserializer_callable is None:
            warnings.warn(f"Unsupported {self.mimetype} mimetype")
            return value

        try:
            return self.deserializer_callable(value)
        except (ValueError, TypeError, AttributeError):
            raise MediaTypeDeserializeError(self.mimetype, value)



import warnings
from typing import Dict
from typing import Optional

from openapi_core.deserializing.media_types.datatypes import (
    DeserializerCallable,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.deserializers import (
    CallableMediaTypeDeserializer,
)


class MediaTypeDeserializersFactory:
    def __init__(
        self,
        media_type_deserializers: Optional[MediaTypeDeserializersDict] = None,
        custom_deserializers: Optional[MediaTypeDeserializersDict] = None,
    ):
        if media_type_deserializers is None:
            media_type_deserializers = {}
        self.media_type_deserializers = media_type_deserializers
        if custom_deserializers is None:
            custom_deserializers = {}
        else:
            warnings.warn(
                "custom_deserializers is deprecated. "
                "Use extra_media_type_deserializers.",
                DeprecationWarning,
            )
        self.custom_deserializers = custom_deserializers

    def create(
        self,
        mimetype: str,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
    ) -> CallableMediaTypeDeserializer:
        if extra_media_type_deserializers is None:
            extra_media_type_deserializers = {}
        deserialize_callable = self.get_deserializer_callable(
            mimetype,
            extra_media_type_deserializers=extra_media_type_deserializers,
        )

        return CallableMediaTypeDeserializer(mimetype, deserialize_callable)

    def get_deserializer_callable(
        self,
        mimetype: str,
        extra_media_type_deserializers: MediaTypeDeserializersDict,
    ) -> Optional[DeserializerCallable]:
        if mimetype in self.custom_deserializers:
            return self.custom_deserializers[mimetype]
        if mimetype in extra_media_type_deserializers:
            return extra_media_type_deserializers[mimetype]
        return self.media_type_deserializers.get(mimetype)



"""OpenAPI core validation processors module"""
from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.spec import Spec
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.proxies import (
    SpecRequestValidatorProxy,
)
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.response.proxies import (
    SpecResponseValidatorProxy,
)


class OpenAPISpecProcessor:
    def __init__(
        self,
        request_unmarshaller: SpecRequestValidatorProxy,
        response_unmarshaller: SpecResponseValidatorProxy,
    ):
        self.request_unmarshaller = request_unmarshaller
        self.response_unmarshaller = response_unmarshaller

    def process_request(
        self, spec: Spec, request: Request
    ) -> RequestUnmarshalResult:
        return self.request_unmarshaller.validate(spec, request)

    def process_response(
        self, spec: Spec, request: Request, response: Response
    ) -> ResponseUnmarshalResult:
        return self.response_unmarshaller.validate(spec, request, response)



"""OpenAPI core validation request protocols module"""
import sys
from typing import Iterator
from typing import Optional

if sys.version_info >= (3, 8):
    from typing import Protocol
    from typing import runtime_checkable
else:
    from typing_extensions import Protocol
    from typing_extensions import runtime_checkable

from openapi_core.protocols import Request
from openapi_core.protocols import WebhookRequest
from openapi_core.spec import Spec


@runtime_checkable
class RequestValidator(Protocol):
    def __init__(self, spec: Spec, base_url: Optional[str] = None):
        ...

    def iter_errors(
        self,
        request: Request,
    ) -> Iterator[Exception]:
        ...

    def validate(
        self,
        request: Request,
    ) -> None:
        ...


@runtime_checkable
class WebhookRequestValidator(Protocol):
    def __init__(self, spec: Spec, base_url: Optional[str] = None):
        ...

    def iter_errors(
        self,
        request: WebhookRequest,
    ) -> Iterator[Exception]:
        ...

    def validate(
        self,
        request: WebhookRequest,
    ) -> None:
        ...



"""OpenAPI core unmarshalling processors module"""
from typing import Optional
from typing import Type

from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.shortcuts import get_classes
from openapi_core.spec import Spec
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType


class UnmarshallingProcessor:
    def __init__(
        self,
        spec: Spec,
        request_unmarshaller_cls: Optional[RequestUnmarshallerType] = None,
        response_unmarshaller_cls: Optional[ResponseUnmarshallerType] = None,
    ):
        self.spec = spec
        if (
            request_unmarshaller_cls is None
            or response_unmarshaller_cls is None
        ):
            classes = get_classes(self.spec)
            if request_unmarshaller_cls is None:
                request_unmarshaller_cls = classes.request_unmarshaller_cls
            if response_unmarshaller_cls is None:
                response_unmarshaller_cls = classes.response_unmarshaller_cls
        self.request_unmarshaller = request_unmarshaller_cls(self.spec)
        self.response_unmarshaller = response_unmarshaller_cls(self.spec)

    def process_request(self, request: Request) -> RequestUnmarshalResult:
        return self.request_unmarshaller.unmarshal(request)

    def process_response(
        self, request: Request, response: Response
    ) -> ResponseUnmarshalResult:
        return self.response_unmarshaller.unmarshal(request, response)



"""OpenAPI core contrib falcon responses module"""
from json import dumps
from typing import Any
from typing import Dict
from typing import Optional

from falcon.request import Request
from falcon.request import RequestOptions
from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict

from openapi_core.datatypes import RequestParameters


class FalconOpenAPIRequest:
    def __init__(
        self,
        request: Request,
        default_when_empty: Optional[Dict[Any, Any]] = None,
    ):
        if not isinstance(request, Request):
            raise TypeError(f"'request' argument is not type of {Request}")
        self.request = request
        if default_when_empty is None:
            default_when_empty = {}
        self.default_when_empty = default_when_empty

        # Path gets deduced by path finder against spec
        self.parameters = RequestParameters(
            query=ImmutableMultiDict(list(self.request.params.items())),
            header=Headers(self.request.headers),
            cookie=self.request.cookies,
        )

    @property
    def host_url(self) -> str:
        assert isinstance(self.request.prefix, str)
        return self.request.prefix

    @property
    def path(self) -> str:
        assert isinstance(self.request.path, str)
        return self.request.path

    @property
    def method(self) -> str:
        assert isinstance(self.request.method, str)
        return self.request.method.lower()

    @property
    def body(self) -> Optional[str]:
        media = self.request.get_media(
            default_when_empty=self.default_when_empty
        )
        # Support falcon-jsonify.
        return dumps(getattr(self.request, "json", media))

    @property
    def mimetype(self) -> str:
        if self.request.content_type:
            assert isinstance(self.request.content_type, str)
            return self.request.content_type.partition(";")[0]

        assert isinstance(self.request.options, RequestOptions)
        assert isinstance(self.request.options.default_media_type, str)
        return self.request.options.default_media_type



"""OpenAPI core validation request module"""
from openapi_core.validation.request.validators import V30RequestBodyValidator
from openapi_core.validation.request.validators import (
    V30RequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V30RequestSecurityValidator,
)
from openapi_core.validation.request.validators import V30RequestValidator
from openapi_core.validation.request.validators import V31RequestBodyValidator
from openapi_core.validation.request.validators import (
    V31RequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V31RequestSecurityValidator,
)
from openapi_core.validation.request.validators import V31RequestValidator
from openapi_core.validation.request.validators import (
    V31WebhookRequestBodyValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestSecurityValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestValidator,
)

__all__ = [
    "V30RequestBodyValidator",
    "V30RequestParametersValidator",
    "V30RequestSecurityValidator",
    "V30RequestValidator",
    "V31RequestBodyValidator",
    "V31RequestParametersValidator",
    "V31RequestSecurityValidator",
    "V31RequestValidator",
    "V31WebhookRequestBodyValidator",
    "V31WebhookRequestParametersValidator",
    "V31WebhookRequestSecurityValidator",
    "V31WebhookRequestValidator",
    "V3RequestValidator",
    "V3WebhookRequestValidator",
]

# alias to the latest v3 version
V3RequestValidator = V31RequestValidator
V3WebhookRequestValidator = V31WebhookRequestValidator



"""OpenAPI core contrib django middlewares module"""
from typing import Callable

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import JsonResponse
from django.http.request import HttpRequest
from django.http.response import HttpResponse

from openapi_core.contrib.django.handlers import DjangoOpenAPIErrorsHandler
from openapi_core.contrib.django.requests import DjangoOpenAPIRequest
from openapi_core.contrib.django.responses import DjangoOpenAPIResponse
from openapi_core.unmarshalling.processors import UnmarshallingProcessor
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)


class DjangoOpenAPIMiddleware:
    request_class = DjangoOpenAPIRequest
    response_class = DjangoOpenAPIResponse
    errors_handler = DjangoOpenAPIErrorsHandler()

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]):
        self.get_response = get_response

        if not hasattr(settings, "OPENAPI_SPEC"):
            raise ImproperlyConfigured("OPENAPI_SPEC not defined in settings")

        self.processor = UnmarshallingProcessor(settings.OPENAPI_SPEC)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        openapi_request = self._get_openapi_request(request)
        req_result = self.processor.process_request(openapi_request)
        if req_result.errors:
            response = self._handle_request_errors(req_result, request)
        else:
            request.openapi = req_result
            response = self.get_response(request)

        openapi_response = self._get_openapi_response(response)
        resp_result = self.processor.process_response(
            openapi_request, openapi_response
        )
        if resp_result.errors:
            return self._handle_response_errors(resp_result, request, response)

        return response

    def _handle_request_errors(
        self, request_result: RequestUnmarshalResult, req: HttpRequest
    ) -> JsonResponse:
        return self.errors_handler.handle(request_result.errors, req, None)

    def _handle_response_errors(
        self,
        response_result: ResponseUnmarshalResult,
        req: HttpRequest,
        resp: HttpResponse,
    ) -> JsonResponse:
        return self.errors_handler.handle(response_result.errors, req, resp)

    def _get_openapi_request(
        self, request: HttpRequest
    ) -> DjangoOpenAPIRequest:
        return self.request_class(request)

    def _get_openapi_response(
        self, response: HttpResponse
    ) -> DjangoOpenAPIResponse:
        return self.response_class(response)



"""OpenAPI core templating paths finders module"""
from typing import Iterator
from typing import List
from typing import Optional
from urllib.parse import urljoin
from urllib.parse import urlparse

from more_itertools import peekable

from openapi_core.schema.servers import is_absolute
from openapi_core.spec import Spec
from openapi_core.templating.datatypes import TemplateResult
from openapi_core.templating.paths.datatypes import Path
from openapi_core.templating.paths.datatypes import PathOperation
from openapi_core.templating.paths.datatypes import PathOperationServer
from openapi_core.templating.paths.exceptions import OperationNotFound
from openapi_core.templating.paths.exceptions import PathNotFound
from openapi_core.templating.paths.exceptions import ServerNotFound
from openapi_core.templating.paths.util import template_path_len
from openapi_core.templating.util import parse
from openapi_core.templating.util import search


class BasePathFinder:
    def __init__(self, spec: Spec, base_url: Optional[str] = None):
        self.spec = spec
        self.base_url = base_url

    def find(self, method: str, name: str) -> PathOperationServer:
        paths_iter = self._get_paths_iter(name)
        paths_iter_peek = peekable(paths_iter)

        if not paths_iter_peek:
            raise PathNotFound(name)

        operations_iter = self._get_operations_iter(method, paths_iter_peek)
        operations_iter_peek = peekable(operations_iter)

        if not operations_iter_peek:
            raise OperationNotFound(name, method)

        servers_iter = self._get_servers_iter(
            name,
            operations_iter_peek,
        )

        try:
            return next(servers_iter)
        except StopIteration:
            raise ServerNotFound(name)

    def _get_paths_iter(self, name: str) -> Iterator[Path]:
        raise NotImplementedError

    def _get_operations_iter(
        self, method: str, paths_iter: Iterator[Path]
    ) -> Iterator[PathOperation]:
        for path, path_result in paths_iter:
            if method not in path:
                continue
            operation = path / method
            yield PathOperation(path, operation, path_result)

    def _get_servers_iter(
        self, name: str, operations_iter: Iterator[PathOperation]
    ) -> Iterator[PathOperationServer]:
        raise NotImplementedError


class APICallPathFinder(BasePathFinder):
    def __init__(self, spec: Spec, base_url: Optional[str] = None):
        self.spec = spec
        self.base_url = base_url

    def _get_paths_iter(self, name: str) -> Iterator[Path]:
        template_paths: List[Path] = []
        paths = self.spec / "paths"
        for path_pattern, path in list(paths.items()):
            # simple path.
            # Return right away since it is always the most concrete
            if name.endswith(path_pattern):
                path_result = TemplateResult(path_pattern, {})
                yield Path(path, path_result)
            # template path
            else:
                result = search(path_pattern, name)
                if result:
                    path_result = TemplateResult(path_pattern, result.named)
                    template_paths.append(Path(path, path_result))

        # Fewer variables -> more concrete path
        yield from sorted(template_paths, key=template_path_len)

    def _get_servers_iter(
        self, name: str, operations_iter: Iterator[PathOperation]
    ) -> Iterator[PathOperationServer]:
        for path, operation, path_result in operations_iter:
            servers = (
                path.get("servers", None)
                or operation.get("servers", None)
                or self.spec.get("servers", [{"url": "/"}])
            )
            for server in servers:
                server_url_pattern = name.rsplit(path_result.resolved, 1)[0]
                server_url = server["url"]
                if not is_absolute(server_url):
                    # relative to absolute url
                    if self.base_url is not None:
                        server_url = urljoin(self.base_url, server["url"])
                    # if no base url check only path part
                    else:
                        server_url_pattern = urlparse(server_url_pattern).path
                if server_url.endswith("/"):
                    server_url = server_url[:-1]
                # simple path
                if server_url_pattern == server_url:
                    server_result = TemplateResult(server["url"], {})
                    yield PathOperationServer(
                        path,
                        operation,
                        server,
                        path_result,
                        server_result,
                    )
                # template path
                else:
                    result = parse(server["url"], server_url_pattern)
                    if result:
                        server_result = TemplateResult(
                            server["url"], result.named
                        )
                        yield PathOperationServer(
                            path,
                            operation,
                            server,
                            path_result,
                            server_result,
                        )


class WebhookPathFinder(BasePathFinder):
    def _get_paths_iter(self, name: str) -> Iterator[Path]:
        if "webhooks" not in self.spec:
            raise PathNotFound("Webhooks not found")
        webhooks = self.spec / "webhooks"
        for webhook_name, path in list(webhooks.items()):
            if name == webhook_name:
                path_result = TemplateResult(webhook_name, {})
                yield Path(path, path_result)

    def _get_servers_iter(
        self, name: str, operations_iter: Iterator[PathOperation]
    ) -> Iterator[PathOperationServer]:
        for path, operation, path_result in operations_iter:
            yield PathOperationServer(
                path,
                operation,
                None,
                path_result,
                {},
            )



from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import List

from openapi_core.casting.schemas.datatypes import CasterCallable
from openapi_core.casting.schemas.exceptions import CastError
from openapi_core.spec import Spec

if TYPE_CHECKING:
    from openapi_core.casting.schemas.factories import SchemaCastersFactory


class BaseSchemaCaster:
    def __init__(self, schema: Spec):
        self.schema = schema

    def __call__(self, value: Any) -> Any:
        if value is None:
            return value

        return self.cast(value)

    def cast(self, value: Any) -> Any:
        raise NotImplementedError


class CallableSchemaCaster(BaseSchemaCaster):
    def __init__(self, schema: Spec, caster_callable: CasterCallable):
        super().__init__(schema)
        self.caster_callable = caster_callable

    def cast(self, value: Any) -> Any:
        try:
            return self.caster_callable(value)
        except (ValueError, TypeError):
            raise CastError(value, self.schema["type"])


class DummyCaster(BaseSchemaCaster):
    def cast(self, value: Any) -> Any:
        return value


class ComplexCaster(BaseSchemaCaster):
    def __init__(self, schema: Spec, casters_factory: "SchemaCastersFactory"):
        super().__init__(schema)
        self.casters_factory = casters_factory


class ArrayCaster(ComplexCaster):
    @property
    def items_caster(self) -> BaseSchemaCaster:
        return self.casters_factory.create(self.schema / "items")

    def cast(self, value: Any) -> List[Any]:
        # str and bytes are not arrays according to the OpenAPI spec
        if isinstance(value, (str, bytes)):
            raise CastError(value, self.schema["type"])

        try:
            return list(map(self.items_caster, value))
        except (ValueError, TypeError):
            raise CastError(value, self.schema["type"])



import warnings
from typing import Any
from typing import Dict
from typing import Hashable
from typing import Mapping
from typing import Optional
from typing import Type
from typing import TypeVar

from jsonschema_spec import Spec as JsonschemaSpec
from jsonschema_spec import default_handlers
from openapi_spec_validator.validation import openapi_spec_validator_proxy
from openapi_spec_validator.validation.protocols import SupportsValidation

TSpec = TypeVar("TSpec", bound="Spec")

SPEC_SEPARATOR = "#"


class Spec(JsonschemaSpec):
    @classmethod
    def create(
        cls: Type[TSpec],
        data: Mapping[Hashable, Any],
        *args: Any,
        url: str = "",
        ref_resolver_handlers: Dict[str, Any] = default_handlers,
        separator: str = SPEC_SEPARATOR,
        validator: Optional[SupportsValidation] = openapi_spec_validator_proxy,
    ) -> TSpec:
        warnings.warn(
            "Spec.create method is deprecated. Use Spec.from_dict instead.",
            DeprecationWarning,
        )

        return cls.from_dict(
            data,
            *args,
            spec_url=url,
            ref_resolver_handlers=ref_resolver_handlers,
            separator=separator,
            validator=validator,
        )

    @classmethod
    def from_dict(
        cls: Type[TSpec],
        data: Mapping[Hashable, Any],
        *args: Any,
        spec_url: str = "",
        ref_resolver_handlers: Mapping[str, Any] = default_handlers,
        separator: str = SPEC_SEPARATOR,
        validator: Optional[SupportsValidation] = openapi_spec_validator_proxy,
    ) -> TSpec:
        if validator is not None:
            validator.validate(data, spec_url=spec_url)

        return super().from_dict(
            data,
            *args,
            spec_url=spec_url,
            ref_resolver_handlers=ref_resolver_handlers,
            separator=separator,
        )



"""OpenAPI core templating media types finders module"""
import fnmatch

from openapi_core.spec import Spec
from openapi_core.templating.media_types.datatypes import MediaType
from openapi_core.templating.media_types.exceptions import MediaTypeNotFound


class MediaTypeFinder:
    def __init__(self, content: Spec):
        self.content = content

    def find(self, mimetype: str) -> MediaType:
        if mimetype in self.content:
            return MediaType(self.content / mimetype, mimetype)

        if mimetype:
            for key, value in self.content.items():
                if fnmatch.fnmatch(mimetype, key):
                    return MediaType(value, key)

        raise MediaTypeNotFound(mimetype, list(self.content.keys()))



import sys
from typing import Any
from typing import List

if sys.version_info >= (3, 8):
    from typing import Protocol
    from typing import runtime_checkable
else:
    from typing_extensions import Protocol
    from typing_extensions import runtime_checkable


@runtime_checkable
class SuportsGetAll(Protocol):
    def getall(self, name: str) -> List[Any]:
        ...


@runtime_checkable
class SuportsGetList(Protocol):
    def getlist(self, name: str) -> List[Any]:
        ...



from openapi_core.contrib.flask.requests import FlaskOpenAPIRequest
from openapi_core.contrib.flask.responses import FlaskOpenAPIResponse

__all__ = [
    "FlaskOpenAPIRequest",
    "FlaskOpenAPIResponse",
]



from email.parser import Parser
from typing import Any
from typing import Dict
from typing import Union
from urllib.parse import parse_qsl


def urlencoded_form_loads(value: Any) -> Dict[str, Any]:
    return dict(parse_qsl(value))


def data_form_loads(value: Union[str, bytes]) -> Dict[str, Any]:
    if isinstance(value, bytes):
        value = value.decode("ASCII", errors="surrogateescape")
    parser = Parser()
    parts = parser.parsestr(value, headersonly=False)
    return {
        part.get_param("name", header="content-disposition"): part.get_payload(
            decode=True
        )
        for part in parts.get_payload()
    }
